#include <linux/module.h>
#include <linux/sched.h>

#include "gmkrc.h"
#include "gmkrc_vmaspy.h"
#include "gmkrc_internal.h"
#include "gmkrc_regcache.h"
#include "gmkrc_messages.h"


gm_status_t
gmkrc_open(struct gm_port **p,
	   unsigned int unit,
	   unsigned int port,
	   const char *port_name,
	   enum gm_api_version version)
{
  gm_status_t status;

  status = gm_open(p, unit, port, port_name, version);
  if ( status != GM_SUCCESS )
    goto error;

  LOG(3, "Port %d on board %d open as %p\n", port, unit, *p);

  BUG_ON(gmkrc_ports[port].port != NULL);
  gmkrc_ports[port].port = *p;

#ifdef CONFIG_GMKRC_REGCACHE
  status = gmkrc_spy_init(&gmkrc_ports[port]);
  if ( status != GM_SUCCESS )
    goto error_with_port;
#endif /* ~CONFIG_GMKRC_REGCACHE */

  return GM_SUCCESS;

#ifdef CONFIG_GMKRC_REGCACHE
 error_with_port:
  gm_close(*p);
#endif /* ~CONFIG_GMKRC_REGCACHE */
 error:
  LOG(3, "Port %d on board %d open failed with error %s\n",
      port, unit, gm_strerror(status));
  return status;
}

void
gmkrc_close(struct gm_port *p)
{
  unsigned int id = p->id;

  LOG(3, "Closing port %p\n", p);

#ifdef CONFIG_GMKRC_REGCACHE
  gmkrc_spy_exit(&gmkrc_ports[id]);
#endif /* ~CONFIG_GMKRC_REGCACHE */
  BUG_ON(gmkrc_ports[id].port == NULL);
  gmkrc_ports[id].port = NULL;
  gm_close(p);
}

gm_status_t
gmkrc_register_memory(struct gm_port *port,
		      void *buffer,
		      unsigned long length)
{
#ifdef CONFIG_GMKRC_REGCACHE

  void *addr = (void*)((unsigned long)buffer & PAGE_MASK);
  unsigned long offset = (unsigned long)buffer & ~PAGE_MASK;
  unsigned long pages = (length+offset + PAGE_SIZE-1) >> PAGE_SHIFT;
  void *mm = NULL;
  unsigned int id = port->id;

  if ( (unsigned long)buffer < PAGE_OFFSET )
    mm = current->mm;

  LOG(3, "Registering buffer %p len %ld in mm %p for port %p\n",
      buffer, length, mm, port);

  gmkrc_register_vma_spy(&gmkrc_ports[id],
			 addr, pages << PAGE_SHIFT);

  return gmkrc_regcache_register(&gmkrc_ports[id],
				 addr, mm, pages);

#else /* ~CONFIG_GMKRC_REGCACHE */

  gm_up_t uptr = GM_PTR_TO_UP (buffer);
  gm_up_t pvma = uptr;

  if ( (unsigned long)buffer < PAGE_OFFSET )
    pvma |= GM_PTR_TO_UP (current->mm) << 32;

  LOG(3, "Registering buffer %llx as %llx len %ld for port %p\n",
      uptr, pvma, length, port);

  return gm_register_memory_ex_up(port, uptr, length, pvma);

#endif /* ~CONFIG_GMKRC_REGCACHE */
}


gm_status_t
gmkrc_deregister_memory(struct gm_port *port,
			void *buffer,
			unsigned long length)
{
#ifdef CONFIG_GMKRC_REGCACHE

  void *addr = (void*)((unsigned long)buffer & PAGE_MASK);
  unsigned long offset = (unsigned long)buffer & ~PAGE_MASK;
  unsigned long pages = (length+offset + PAGE_SIZE-1) >> PAGE_SHIFT;
  void *mm = NULL;
  unsigned int id = port->id;

  if ( (unsigned long)buffer < PAGE_OFFSET )
    mm = current->mm;

  LOG(3, "Deregistering buffer %p len %ld in mm %p for port %p\n",
      buffer, length, mm, port);

  return gmkrc_regcache_deregister(&gmkrc_ports[id],
				   addr, mm, pages);

#else /* ~CONFIG_GMKRC_REGCACHE */

  gm_up_t uptr = GM_PTR_TO_UP (buffer);
  gm_up_t pvma = uptr;

  if ( (unsigned long)buffer < PAGE_OFFSET )
    pvma |= GM_PTR_TO_UP (current->mm) << 32;

  LOG(3, "Deregistering buffer as %llx len %ld for port %p\n",
      pvma, length, port);

  return gm_deregister_memory_up(port, pvma, length);

#endif /* ~CONFIG_GMKRC_REGCACHE */
}


void
gmkrc_send_with_callback(struct gm_port *port,
			 void *message,
			 unsigned int size,
			 gm_size_t len,
			 unsigned int priority,
			 unsigned int target_node_id,
			 unsigned int target_port_id,
			 gm_send_completion_callback_t callback,
			 void *context)
{
  gm_up_t uptr = GM_PTR_TO_UP (message);
  gm_up_t pvma = uptr;

  if ( (unsigned long)message < PAGE_OFFSET )
    pvma |= GM_PTR_TO_UP (current->mm) << 32;

  LOG(3, "Sending buffer %llx size %d len %ld prio %d target %d port %d with port %p\n",
      pvma, size, len, priority, target_node_id, target_port_id, port);

  return gm_send_with_callback_up(port, pvma, size, len, priority,
			          target_node_id, target_port_id, callback, context);
}


void
gmkrc_provide_receive_buffer_with_tag(struct gm_port *port,
				      void *buffer,
				      unsigned int size,
				      unsigned int priority,
				      unsigned int tag)
{
  gm_up_t uptr = GM_PTR_TO_UP (buffer);
  gm_up_t pvma = uptr;

  if ( (unsigned long)buffer < PAGE_OFFSET )
    pvma |= GM_PTR_TO_UP (current->mm) << 32;

  LOG(3, "Providing buffer %llx size %d prio %d for port %p\n",
      pvma, size, priority, port);

  return gm_provide_receive_buffer_with_tag_up(port, pvma, size, priority, tag);
}
