#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/slab.h>

#include "gmkrc.h"
#include "gmkrc_internal.h"
#include "gmkrc_messages.h"


/*********
 * Entry *
 *********/

struct gmkrc_entry {
  void *addr;
  void *mm; // vma ?
  struct gm_port *port;
  unsigned int count;
  struct list_head hash;
  struct list_head lru;
};


/* LRU at the begin */
static struct list_head gmkrc_lru_head;


/* Hash Table */
#define HASH_TABLE_BITS 12
#define HASH_TABLE_SIZE (1<<HASH_TABLE_BITS)

static struct list_head gmkrc_entry_hash_head[HASH_TABLE_SIZE];


/* Entry cache */
static kmem_cache_t *gmkrc_entry_cachep;


/******************
 * Hash functions *
 ******************/

static inline
unsigned int gmkrc_entry_hash_func(struct gm_port *port, void *addr, void *mm)
{
  return (unsigned int)(((unsigned long)addr+(unsigned long)mm+(unsigned long)port)>>(PAGE_SHIFT+HASH_TABLE_BITS));
}

static inline
struct gmkrc_entry *gmkrc_entry_hash_find(struct gm_port *port, void *addr, void *mm)
{
  unsigned int key = gmkrc_entry_hash_func(port, addr, mm);
  struct list_head *tmp;
  struct gmkrc_entry *entry;

  list_for_each(tmp, &gmkrc_entry_hash_head[key]) {
    entry = list_entry(tmp, struct gmkrc_entry, hash);
    if ( entry->addr == addr && entry->mm == mm && entry->port == port )
      return entry;
  }

  return NULL;    
}

/***************
 * Entry Cache *
 ***************/

static inline
struct gmkrc_entry *gmkrc_entry_cache_alloc(void *addr, void *mm, struct gm_port *port)
{
  struct gmkrc_entry *entry;
  unsigned int key;

  entry = (struct gmkrc_entry*) kmem_cache_alloc(gmkrc_entry_cachep, SLAB_KERNEL);
  if ( !entry ) {
    ERROR("cannot allocate entry\n");
    return NULL;
  }

  entry->addr = addr;
  entry->mm = mm;
  entry->port = port;
  entry->count = 1;
  
  /* add to LRU */
  list_add(&entry->lru, &gmkrc_lru_head);
  /* hash */
  key = gmkrc_entry_hash_func(port, addr, mm);
  list_add_tail(&entry->hash, &gmkrc_entry_hash_head[key]);

  return entry;
}

static inline
void gmkrc_entry_cache_free(struct gmkrc_entry *entry)
{
  __list_del(entry->hash.prev, entry->hash.next);
  __list_del(entry->lru.prev, entry->lru.next);
  kmem_cache_free(gmkrc_entry_cachep, entry);
}

/*****************
 * LRU functions *
 *****************/

static inline
void gmkrc_lru_use(struct gmkrc_entry *entry)
{
  list_move(&entry->lru, &gmkrc_lru_head);
}

/*************
 * Init/Exit *
 *************/

int gmkrc_regcache_init(void)
{
  int error;
  int i;

  /* initialize the entry cache */
  error = -ENOMEM;
  gmkrc_entry_cachep = kmem_cache_create("gmkrc_entry", 
					 sizeof(struct gmkrc_entry),
					 0, 0, NULL, NULL);
  if ( !gmkrc_entry_cachep ) {
    ERROR("cannot allocate entry cache\n");
    goto error;
  }

  /* initialize lru list head */
  INIT_LIST_HEAD(&gmkrc_lru_head);
  /* initialize hash table heads */
  for(i=0; i<HASH_TABLE_SIZE; i++)
    INIT_LIST_HEAD(&gmkrc_entry_hash_head[i]);

  return 0;

 error:
  return error;
}

void gmkrc_regcache_exit(void)
{
  struct list_head *tmp, *next;
  struct gmkrc_entry *entry;

  /* destroy all entries using the lru list */
  list_for_each_safe(tmp, next, &gmkrc_lru_head) {
    __list_del(tmp->prev, tmp->next);
    entry = list_entry(tmp, struct gmkrc_entry, hash);
    gmkrc_entry_cache_free(entry);
    // we should really deregister here...
    // GM will do that on gm_close
  }

  /* destroy the entry cache */
  kmem_cache_destroy(gmkrc_entry_cachep);
}

/****************
 * Registration *
 ****************/

#ifdef CONFIG_GMKRC_AMOUNT
static unsigned long gmkrc_registrable_pages = CONFIG_GMKRC_AMOUNT_LEVEL;
#endif


static inline
gm_status_t gmkrc_real_register(struct gm_port *port,
				void *addr,
				void *mm,
				unsigned long pages)
{
  gm_status_t status;
  gm_up_t uptr = GM_PTR_TO_UP (addr);
  gm_up_t pvma = uptr | (GM_PTR_TO_UP (mm) << 32);
  gm_size_t length = pages<<PAGE_SHIFT;

#ifdef CONFIG_GMKRC_AMOUNT
  if ( gmkrc_registrable_pages < pages ) {
    LOG(3, "Not enough pages registrable (%ld/%ld)\n",
	gmkrc_registrable_pages, pages);
    return GM_FAILURE;
  }
#endif

  status = gm_register_memory_ex_up(port, uptr, length, pvma);

#ifdef CONFIG_GMKRC_AMOUNT
  if ( status == GM_SUCCESS )
    gmkrc_registrable_pages -= pages;
#endif

  return status;
}

static inline
gm_status_t gmkrc_real_deregister(struct gm_port *port,
				  void *addr,
				  void *mm,
				  unsigned long pages)
{
  gm_status_t status;
  gm_up_t uptr = GM_PTR_TO_UP (addr);
  gm_up_t pvma = uptr | (GM_PTR_TO_UP (mm) << 32);

  status = gm_deregister_memory_up(port, pvma, pages<<PAGE_SHIFT);

#ifdef CONFIG_GMKRC_AMOUNT
  if ( status == GM_SUCCESS )
    gmkrc_registrable_pages += pages;
#endif

  return status;
}

/************
 * RegCache *
 ************/

static
gm_status_t garbage_collecting(unsigned long required_pages)
{
  struct gmkrc_entry *entry;
  struct list_head *head = gmkrc_lru_head.prev;

  void *current_addr = NULL;
  void *current_mm = NULL;
  struct gm_port *current_port = NULL;
  unsigned long current_pages = 0;

  LOG(2, "Trying to garbage %ld pages\n", required_pages);

  do {
    entry = list_entry(head, struct gmkrc_entry, lru);
    LOG(5, "Dealing with entry at %p in %p for port %p with count %d\n",
	entry->addr, entry->mm, entry->port, entry->count);
    
    if ( entry->count ) {
      /* cannot release this entry */
      if ( current_addr ) {
	/* release previous sequence */
	LOG(3, "Garbaging %ld pages at %p in %p for port %p\n",
	    current_pages, current_addr, current_mm, current_port);
	gmkrc_real_deregister(current_port, current_addr, current_mm, current_pages);
	/* no started sequence for now */
	current_addr = NULL;
      }
      
      head = head->prev;
      continue;
    }
    
    /* start a new sequence. this page will be released */
    required_pages--;
    
    if ( !current_addr ) {
      /* start a new sequence */
      LOG(4, "Starting a new sequence\n");
      current_addr = entry->addr;
      current_mm = entry->mm;
      current_port = entry->port;
      current_pages = 1;
    } else if ( entry->addr == current_addr+(current_pages<<PAGE_SHIFT)
		&& entry->mm == current_mm
		&& entry->port == current_port ) {
      /* continue previous sequence */
      LOG(4, "Continuing sequence\n");
      current_pages++;
    } else {
      /* finish previous sequence and start a new one */
      LOG(3, "Garbaging %ld pages at %p in %p for port %p\n",
	  current_pages, current_addr, current_mm, current_port);
      gmkrc_real_deregister(current_port, current_addr, current_mm, current_pages);
      current_addr = entry->addr;
      current_mm = entry->mm;
      current_port = entry->port;
      current_pages = 1;
    }
    
    /* get next entry, that is previous one */
    head = head->prev;    
    
    /* remove the entry */
    gmkrc_entry_cache_free(entry);

    /* repeat until scanned the whole list, or got enough pages */
  } while ( head != &gmkrc_lru_head && required_pages > 0 );
  
  if ( current_addr ) {
    /* release remaining sequence */
    LOG(3, "Garbaging %ld pages at %p in %p for port %p\n",
	current_pages, current_addr, current_mm, current_port);
    gmkrc_real_deregister(current_port, current_addr, current_mm, current_pages);
  }

  LOG(3, "Still %ld pages to garbage\n", required_pages);
  return required_pages ? GM_FAILURE : GM_SUCCESS;
}

static inline
gm_status_t gmkrc_regcache_register_try(struct gm_port *port,
					void *addr,
					void *mm,
					unsigned long pages)
{
  gm_status_t status;

  LOG(2, "Trying to really register %ld pages at %p in %p for port %p\n",
      pages, addr, mm, port);

  status = gmkrc_real_register(port, addr, mm, pages);
  if ( status == GM_SUCCESS )
    return GM_SUCCESS;

  /* register failed */
  status = garbage_collecting(pages);
  if ( status != GM_SUCCESS )
    return status;

  LOG(2, "Second try for really registering %ld pages at %p in %p for port %p\n",
      pages, addr, mm, port);

  /* register should now succeed */
  return gmkrc_real_register(port, addr, mm, pages);
}

/****************
 * Entry Points *
 ****************/

gm_status_t gmkrc_regcache_register(struct gmkrc_port *gmkrc_port,
				    void *addr,
				    void *mm,
				    unsigned long pages)
{
  struct gm_port *port = gmkrc_port->port;
  struct gmkrc_entry *entry;
  gm_status_t status = GM_SUCCESS;
  int i;

  void *current_addr = NULL;
  unsigned long current_pages = 0;

  LOG(1, "Registering %ld pages at %p in %p for port %p\n",
      pages, addr, mm, port);

  down(&gmkrc_port->sem);

  for(i=0; i<pages; i++) {
    entry = gmkrc_entry_hash_find(port, addr, mm);

    if ( entry ) {

      /* this entry exists and is thus already registered */
      entry->count++;
      gmkrc_lru_use(entry);

      if ( current_addr ) {
	/* register previous sequence */
	status = gmkrc_regcache_register_try(port, current_addr, mm, current_pages);
	if ( status != GM_SUCCESS )
	  goto register_failed;
      }

    } else {

      /* allocate and register */
      entry = gmkrc_entry_cache_alloc(addr, current->mm, port);
      if ( !entry )
	goto out_of_mem;

      if ( current_addr ) {
	/* continue the sequence */
	current_pages++;
      } else {
	/* start a new sequence */
	current_addr = entry->addr;
	current_pages = 1;
      }

    }

    addr += PAGE_SIZE;
  }

  if ( current_addr ) {
    /* register last sequence */
    status = gmkrc_regcache_register_try(port, current_addr, mm, current_pages);
    if ( status != GM_SUCCESS )
      goto register_failed;
  }

  up(&gmkrc_port->sem);
  return GM_SUCCESS;
  
 out_of_mem:
  ERROR("Registration failed\n");
  if ( !current_addr )
    goto error;

 register_failed:
  ERROR("Releasing allocated entries\n");
  for(i=0; i<current_pages; i++) {
    entry = gmkrc_entry_hash_find(port, current_addr, mm);
    // entry MUST be != NULL
    
    /* and we remove it because it was not registered */
    gmkrc_entry_cache_free(entry);

    current_addr += PAGE_SIZE;
  }

 error:
  up(&gmkrc_port->sem);
  return status;
}

gm_status_t gmkrc_regcache_deregister(struct gmkrc_port *gmkrc_port,
				      void *addr,
				      void *mm,
				      unsigned long pages)
{
  struct gm_port *port = gmkrc_port->port;
  struct gmkrc_entry *entry;
  int i;

  LOG(1, "Deregistering %ld pages at %p in %p for port %p\n",
      pages, addr, mm, port);

  down(&gmkrc_port->sem);

  for(i=0; i<pages; i++) {
    entry = gmkrc_entry_hash_find(port, addr, mm);
    // entry MUST be != NULL

    entry->count--;

    addr += PAGE_SIZE;
  }

  up(&gmkrc_port->sem);
  return GM_SUCCESS;
}

void gmkrc_regcache_clear(struct gmkrc_port *gmkrc_port,
			  void *addr,
			  void *mm,
			  unsigned long pages)
{
  struct gm_port *port = gmkrc_port->port;
  struct gmkrc_entry *entry;
  int i;
  void *current_addr = NULL;
  unsigned long current_pages = 0;

  LOG(1, "Trying to clear %ld pages at %p in %p for port %p\n",
      pages, addr, mm, port);

  down(&gmkrc_port->sem);

  for(i=0; i<pages; i++) {
    entry = gmkrc_entry_hash_find(port, addr, mm);

    if ( entry ) {

      /* have to deregister this one */
      if ( entry->count )
        ERROR("Clearing non-free regcache entry %p for %p in %p for port %p (count %d)\n",
              entry, addr, mm, port, entry->count);

      if ( current_addr ) {
	/* continue current sequence */
	current_pages++;
      } else {
	/* start a new sequence */
	current_addr = addr;
	current_pages = 1;
      }

      /* release this entry */
      gmkrc_entry_cache_free(entry);      

    } else if ( current_addr ) {

      /* deregister previous sequence */
      LOG(2, "Really deregistering %ld pages at %p in %p for port %p\n",
	  current_pages, current_addr, mm, port);
      gmkrc_real_deregister(port, current_addr, mm, current_pages);
      current_addr = NULL;

    }
    addr += PAGE_SIZE;
  }
  
  if ( current_addr ) {
    /* deregister last sequence */
    LOG(2, "Really deregistering %ld pages at %p in %p for port %p\n",
	current_pages, current_addr, mm, port);
    gmkrc_real_deregister(port, current_addr, mm, current_pages);
  }
  
  up(&gmkrc_port->sem);
}
