#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/mm.h>

#include "gmkrc.h"
#include "gmkrc_internal.h"
#include "gmkrc_messages.h"
#include "gmkrc_regcache.h"

/*********************
 * Vma Spy Callbacks *
 *********************/

static
void gmkrc_notify_unmap(struct vm_area_struct *vma,
			unsigned long start,
			unsigned long end,
			void *data)
{
  unsigned long pages;
  void *addr;
  unsigned long offset;

  LOG(1, "unmap called for vma %p from %lx to %lx with port %p\n",
      vma, start, end, ((struct gmkrc_port *)data)->port);

  offset = start & ~PAGE_MASK;
  addr = (void*)(start & PAGE_MASK);
  pages = (end-start+offset + PAGE_SIZE-1) >> PAGE_SHIFT;

#if CONFIG_GMKRC_REGCACHE
  gmkrc_regcache_clear((struct gmkrc_port *)data,
		       (void*)start, vma->vm_mm, pages);
#else
  ERROR("Unmapping of vma %p in mm %p from %lx to %lx\n",
        vma, vma->vm_mm, start, end);
#endif
}

static
void gmkrc_notify_fork(struct vm_area_struct *vma,
		       void *data)
{
  unsigned long pages;
  void *addr;
  unsigned long offset;

  LOG(1, "fork called for vma %p with port %p\n",
      vma, ((struct gmkrc_port *)data)->port);

  offset = vma->vm_start & ~PAGE_MASK;
  addr = (void*)(vma->vm_start & PAGE_MASK);
  pages = (vma->vm_end-vma->vm_start+offset + PAGE_SIZE-1) >> PAGE_SHIFT;

#if CONFIG_GMKRC_REGCACHE
  gmkrc_regcache_clear((struct gmkrc_port *)data,
		       addr, vma->vm_mm, pages);
#else
  ERROR("Duplicating of vma %p in mm %p (%lx to %lx)\n",
        vma, vma->vm_mm, vma->vm_start, vma->vm_end);
#endif
}

/***********
 * Vma Spy *
 ***********/

gm_status_t gmkrc_register_vma_spy(struct gmkrc_port *port, void *addr, unsigned long length)
{
  struct vm_area_struct *vma;
  struct vma_spy_type *type;
  unsigned long current_addr = (unsigned long)addr;
  unsigned long current_length = length;

  LOG(2, "Wrapping from %p to %p in %p for port %p\n", 
      addr, addr+length, current->mm, port->port);

  while ( 1 ) {
    vma = find_vma(current->mm, current_addr);
    if ( !vma )
      return GM_MEMORY_FAULT;
    
    /* is there a spy for this port ? */
    if ( !port->port )
      return GM_INVALID_PARAMETER; // good ?

    /* is this vma spied by this port ? */
    type = &port->vma_spy_type;
    if ( !is_vma_spy_registered(vma, type) ) {
      struct vma_spy *spy = register_vma_spy(vma, type);
      if ( !spy )
	return GM_OUT_OF_MEMORY;
    }

    if ( current_addr + current_length <= vma->vm_end )
      break;
    current_length -= vma->vm_end - current_addr;
    current_addr = vma->vm_end;
  }

  return GM_SUCCESS;
}

/*************
 * Init/Exit *
 *************/

gm_status_t gmkrc_spy_init(struct gmkrc_port *port)
{
  struct vma_spy_type *type = &port->vma_spy_type;

  type->unmap = gmkrc_notify_unmap;
  type->fork = gmkrc_notify_fork;
  type->data = port;
  INIT_LIST_HEAD(&type->vma_list);
  spin_lock_init(&type->vma_lock);

  LOG(1, "Initialized spy type %p for port (%d,%p)\n",
      type, port->port->id, port->port);

  return GM_SUCCESS;
}

void gmkrc_spy_exit(struct gmkrc_port *port)
{
  struct vma_spy_type *type = &port->vma_spy_type;

  LOG(1, "Removing spy type %p for port (%d,%p)\n",
      type, port->port->id, port->port);

  unregister_vma_spy_type(type);
}

void gmkrc_vma_spies_exit(void)
{
  int i;

  for(i=0; i<GM_NUM_PORTS; i++)
    if ( gmkrc_ports[i].port != NULL )
      unregister_vma_spy_type(&gmkrc_ports[i].vma_spy_type);
}
